@@ -45,7 +45,7 @@ xdg_surface_handle_configure (void *data,
     (void)_xdg_surface;
 
     xdg_surface_ack_configure (self->xdg_surface, serial);
-    self->super.configured = TRUE;
+    self->super.awaiting_configure = FALSE;
 }
 
 static const struct xdg_surface_listener xdg_surface_listener = {
@@ -153,6 +153,16 @@ xdg_popup_surface_map (CustomShellSurface *super, struct wl_surface *wl_surface)
 
     g_return_if_fail (!self->xdg_popup);
     g_return_if_fail (!self->xdg_surface);
+    g_return_if_fail (self->position.transient_for_shell_surface);
+    GtkWindow *parent_gtk_window = custom_shell_surface_get_gtk_window (self->position.transient_for_shell_surface);
+    g_return_if_fail (parent_gtk_window);
+    GdkWindow *parent_gdk_window = gtk_widget_get_window (GTK_WIDGET (parent_gtk_window));
+    g_return_if_fail (parent_gdk_window);
+    if (!gdk_wayland_window_get_wl_surface (parent_gdk_window)) {
+        // See https://github.com/wmww/gtk-layer-shell/issues/207
+        g_warning ("Failed to create popup because parent is not mapped");
+        return;
+    }
 
     GtkWindow *gtk_window = custom_shell_surface_get_gtk_window (super);
     GdkWindow *gdk_window = gtk_widget_get_window (GTK_WIDGET (gtk_window));
@@ -178,6 +188,7 @@ xdg_popup_surface_map (CustomShellSurface *super, struct wl_surface *wl_surface)
     self->xdg_surface = xdg_wm_base_get_xdg_surface (xdg_wm_base_global, wl_surface);
     g_return_if_fail (self->xdg_surface);
     xdg_surface_add_listener (self->xdg_surface, &xdg_surface_listener, self);
+    super->awaiting_configure = TRUE;
 
     CustomShellSurface *transient_for_shell_surface = self->position.transient_for_shell_surface;
     self->xdg_popup = custom_shell_surface_add_popup (transient_for_shell_surface,
